{
 "metadata": {
  "name": "",
  "signature": "sha256:700771a6ff8ca29a171c4fc2b5ec08ab4b0c14551c10f96cb2bd5d17505025c0"
 },
 "nbformat": 3,
 "nbformat_minor": 0,
 "worksheets": [
  {
   "cells": [
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "# Virtual Lab 3 Solution: The Propose & Reject Algorithm\n",
      "\n",
      "### EECS 70: Discrete Mathematics and Probability Theory, Fall 2014\n",
      "\n",
      "#### Due Date: Monday, September 22nd, 2014 at 12pm"
     ]
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "**Instructions**:\n",
      "\n",
      "* Complete this lab by filling in all of the required functions, marked with `\"YOUR CODE HERE\"`\n",
      "* If you plan to use Python, make sure to go over **Tutorial 1A: Introduction to Python and IPython** before attempting the lab\n",
      "* Make sure you run every code cell one after another, i.e. don't skip any cell. A shortcut for doing this in the notebook is Shift+Enter. When you finish, choose 'Cell > Run All' to test your code one last time all at once.\n",
      "* **For the sake of simplicity, assume that men and women are both represented by numerals. Whenever you see \"A\", translate it to 1 and so on.**\n",
      "* Please do not hardcode the result or change any function without the `\"YOUR CODE HERE\"` mark\n",
      "* Questions? Bring them to our Office Hour and/or ask on Piazza\n",
      "* Good luck, and have fun!"
     ]
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "## Table of Contents\n",
      "\n",
      "The number inside parentheses is the number of functions you are required to fill out for this lab. Always make sure to double check before you submit.\n",
      "\n",
      "* [Question 1: Create lists of men and women](#q1) (2)\n",
      "* [Question 2: What if women make the first move?](#q2) (1)\n",
      "* [Question 3: Same final pairings](#q3) (1)\n",
      "* [Question 3: How long does the algorithm take?](#q4) (1)"
     ]
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Below you will find the implementation of a Person class in Python. Since we do not expect you to use Object-oriented Programming (OOP) concepts in EECS 70, don't worry if you have trouble understand every detail of the code.\n",
      "\n",
      "That said, you are still expected to read through the code & the docstring of each function and have a general idea of what each function does. Understanding [abstraction][1] is at the heart of Computer Science.\n",
      "\n",
      "[1]: https://en.wikipedia.org/wiki/Abstraction_(computer_science)"
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "class Person:\n",
      "    \"\"\"\n",
      "    A Person represents one of the participants of the propose & reject algorithm\n",
      "    Instances of Person must be stored in lists called 'men' and 'women'\n",
      "    \"\"\"\n",
      "    \n",
      "    def __init__(self, gender, num, ranks):\n",
      "        \"\"\"\n",
      "        Initializes the person to either a man ('M') or woman ('W'),\n",
      "        and gives him/her an identifying number \n",
      "        and a preference list indicating how they rank the members of the opposite gender\n",
      "        \"\"\"\n",
      "\n",
      "        if gender not in ['M', 'W']:\n",
      "            raise Exception(\"Not male or female\")\n",
      "        self.gender = gender\n",
      "        self.num = num\n",
      "        self.ranks = ranks \n",
      "        self.partner = None\n",
      "    \n",
      "    def get_highest_ranked_person(self):\n",
      "        \"\"\"\n",
      "        Returns the most desirable of this Person's possible matches\n",
      "        \"\"\"\n",
      "\n",
      "        if self.gender == 'M':\n",
      "            return women[self.ranks[0]-1]\n",
      "        else:\n",
      "            return men[self.ranks[0]-1]\n",
      "    \n",
      "    def get_rank(self, other_person):\n",
      "        \"\"\"\n",
      "        Returns the position of \"other_person\" in this Person's ranking list\n",
      "        \"\"\"\n",
      "\n",
      "        if self.gender == other_person.gender:\n",
      "            raise Exception(\"Two people of the same gender don't have a\" +\n",
      "                            \" ranking for each other\")\n",
      "        return self.ranks.index(other_person.num) + 1\n",
      "    \n",
      "    def get_partner(self):\n",
      "        \"\"\"\n",
      "        Returns the current partner if there is any, or None otherwise\n",
      "        \"\"\"\n",
      "\n",
      "        return self.partner\n",
      "    \n",
      "    def maybe(self, other_person):\n",
      "        \"\"\"\n",
      "        Changes this Person and other_person's partners to be each other\n",
      "        \"\"\"\n",
      "        \n",
      "        self.partner = other_person\n",
      "        other_person.partner = self\n",
      "\n",
      "    def remove_from_list(self, other_person):\n",
      "        \"\"\"\n",
      "        Removes other_person from this Person's possible matches\n",
      "        \"\"\"\n",
      "\n",
      "        self.ranks.remove(other_person.num)\n",
      "    \n",
      "    def reset(self):\n",
      "        \"\"\"\n",
      "        Resets the partner of the current Person to None\n",
      "        \"\"\"\n",
      "\n",
      "        self.partner = None\n",
      "        \n",
      "    def __str__(self):\n",
      "        \"\"\"\n",
      "        Prints a Person object\n",
      "        \"\"\"\n",
      "        \n",
      "        return \",\".join([self.gender, str(self.num), str(self.ranks),\n",
      "                         str(None if not self.partner else self.partner.num)])"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 48
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "<a id='q1'></a>\n",
      "## Question 1: Create lists of men and women"
     ]
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "In Python, you can create an instance of a class using the syntax `ClassName(class_params)`.\n",
      "\n",
      "For example, in a 2x2 instance, if we want to create man 1 who prefers woman 2 to woman 1, we could write:\n",
      "\n",
      "`man1 = Person('M', 1, [2, 1])`"
     ]
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Here's an example of how to create lists of men and women using the example from Note 3, page 1."
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "from IPython.display import Image\n",
      "Image(url='http://inst.eecs.berkeley.edu/~cs199-xs/stable.png')"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "html": [
        "<img src=\"http://inst.eecs.berkeley.edu/~cs199-xs/stable.png\"/>"
       ],
       "metadata": {},
       "output_type": "pyout",
       "prompt_number": 49,
       "text": [
        "<IPython.core.display.Image at 0xa4d34cc>"
       ]
      }
     ],
     "prompt_number": 49
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def create_example_men_list():\n",
      "    \"\"\"\n",
      "    This function should return a list of size 3, where each\n",
      "    element in the list is a man from the above example\n",
      "    \"\"\"\n",
      "    \n",
      "    return [Person('M', 1, [1,2,3]), Person('M', 2, [2,1,3]), Person('M', 3, [1,2,3])]"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 50
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def create_example_women_list():\n",
      "    \"\"\"\n",
      "    This function should return a list of size 3, where each\n",
      "    element in the list is a woman from the above example\n",
      "    \"\"\"\n",
      "    \n",
      "    return [Person('W', 1, [2,1,3]), Person('W', 2, [1,2,3]), Person('W', 3, [1,2,3])]"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 51
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Now it's your turn. Using the example from Note 4, page 6, which is shown below again for your convenience, create a list of all the men and another list containing all the women. \n",
      "\n",
      "**Reminder**: we are using numerals to represent both genders."
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "from IPython.display import Image\n",
      "Image(url='http://inst.eecs.berkeley.edu/~cs199-xs/stable2.png')"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "html": [
        "<img src=\"http://inst.eecs.berkeley.edu/~cs199-xs/stable2.png\"/>"
       ],
       "metadata": {},
       "output_type": "pyout",
       "prompt_number": 52,
       "text": [
        "<IPython.core.display.Image at 0xa4d368c>"
       ]
      }
     ],
     "prompt_number": 52
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def create_men_list():\n",
      "    \"\"\"\n",
      "    This function should return a list of size 4, where each\n",
      "    element in the list is a man from the above example\n",
      "    \n",
      "    YOUR CODE HERE\n",
      "    \"\"\"\n",
      "    \n",
      "    return [Person('M', 1, [1,2,3,4]), Person('M', 2, [1,4,3,2]),\n",
      "            Person('M', 3, [1,3,2,4]), Person('M', 4, [1,2,3,4])]"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 53
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def create_women_list():\n",
      "    \"\"\"\n",
      "    This function should return a list of size 4, where each\n",
      "    element in the list is a woman from the above example\n",
      "    \n",
      "    YOUR CODE HERE\n",
      "    \"\"\"\n",
      "    \n",
      "    return [Person('W', 1, [1,3,2,4]), Person('W', 2, [4,3,2,1]),\n",
      "            Person('W', 3, [2,3,1,4]), Person('W', 4, [3,4,2,1])]"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 54
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def check_q1():\n",
      "    \"\"\"\n",
      "    This function checks if the functions you created above\n",
      "    returns two lists of the form we expect, so that you can\n",
      "    move on to the next question. It does not check if the \n",
      "    preference list for each Person is correct, so be careful!\n",
      "    \"\"\"\n",
      "    \n",
      "    men = create_men_list()\n",
      "    women = create_women_list()\n",
      "    if len(men) == 4 and len(women) == 4 and\\\n",
      "       all(isinstance(m, Person) for m in men) and\\\n",
      "       all(isinstance(w, Person) for w in women):\n",
      "        print \"Your lists have the right syntax.\"\n",
      "    else:\n",
      "        print \"Your lists are incorrect. Please doublecheck.\""
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 55
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "check_q1()"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "Your lists have the right syntax.\n"
       ]
      }
     ],
     "prompt_number": 56
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "<a id='q2'></a>\n",
      "## Question 2: What if women make the first move?"
     ]
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Below you will find a simple implementation of the traditional Propose & Reject algorithm, where the men propose first. The final pairing is male-optimal, as we learn in lecture.\n",
      "\n",
      "We'll use the example on page 1 of Note 4, in case you didn't complete Question 1. That said, you should be able to change it to any example you like, including the example on page 6 of Note 4 that you implemented above."
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "# Gets the lists of men and women\n",
      "men = create_example_men_list()\n",
      "women = create_example_women_list()"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 57
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def men_propose():\n",
      "    \"\"\"\n",
      "    Runs the traditional Propose & Reject algorithm where men propose\n",
      "    Returns the number of days the algorithm takes to complete\n",
      "    \"\"\"\n",
      "\n",
      "    num_days = 0\n",
      "    has_rejection = True\n",
      "    while has_rejection:\n",
      "        has_rejection = False\n",
      "        proposals = {}\n",
      "\n",
      "        # initialization\n",
      "        for woman in women:\n",
      "            proposals[woman] = []\n",
      "       \n",
      "        # propose phase\n",
      "        for man in men:\n",
      "            ideal_woman = man.get_highest_ranked_person()\n",
      "            proposals[ideal_woman].append(man)\n",
      "       \n",
      "        # decision phase\n",
      "        for woman in women:\n",
      "            if len(proposals[woman]) == 0:\n",
      "                continue\n",
      "            best = proposals[woman][0]\n",
      "            for proposer in proposals[woman]:\n",
      "                if woman.get_rank(proposer) > woman.get_rank(best):\n",
      "                    proposer.remove_from_list(woman) # reject\n",
      "                    has_rejection = True\n",
      "                if woman.get_rank(proposer) < woman.get_rank(best):\n",
      "                    best.remove_from_list(woman) # reject\n",
      "                    has_rejection = True\n",
      "                    best = proposer\n",
      "            woman.maybe(best)\n",
      "        num_days += 1\n",
      "    return num_days"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 58
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def print_men_pairing():\n",
      "    \"\"\"\n",
      "    Prints out the final pairings where men make the proposal\n",
      "    \"\"\"\n",
      "\n",
      "    print \"Final pairings:\"\n",
      "    print \"(Man, Woman)\"\n",
      "    for man in men:\n",
      "        print \"(%s, %s)\" % (man.num, man.get_partner().num)"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 59
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def get_men_pairing():\n",
      "    \"\"\"\n",
      "    Retrieves the final pairings where men make the proposal\n",
      "    \"\"\"\n",
      "    \n",
      "    pairings = []\n",
      "    for man in men:\n",
      "        pairings.append((man.num, man.get_partner().num))\n",
      "    return pairings"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 60
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "men_propose()\n",
      "men_pairing = get_men_pairing()\n",
      "print_men_pairing()"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "Final pairings:\n",
        "(Man, Woman)\n",
        "(1, 1)\n",
        "(2, 2)\n",
        "(3, 3)\n"
       ]
      }
     ],
     "prompt_number": 61
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "**It's your turn**. Use the code above and make any necessary changes, create a new version to run for the case where women propose first. Please do not hardcode the result, as we'll run your implementation on other instances.\n",
      "\n",
      "*Hint*: A good place to start would be to copy the entire body of `men_propose()` function and reverse the role of men and women. Make sure you understand each part of the algorithm instead of blindly changing the variables' names, however."
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "# Gets the lists of men and women\n",
      "men = create_example_men_list()\n",
      "women = create_example_women_list()"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 62
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def women_propose():\n",
      "    \"\"\"\n",
      "    Runs the flipped Propose & Reject algorithm where women propose\n",
      "    Returns the number of days the algorithm takes to complete\n",
      "    \n",
      "    YOUR CODE HERE\n",
      "    \"\"\"\n",
      "\n",
      "    num_days = 0\n",
      "    has_rejection = True\n",
      "    while has_rejection:\n",
      "        has_rejection = False\n",
      "        proposals = {}\n",
      "\n",
      "        # initialization\n",
      "        for man in men:\n",
      "            proposals[man] = []\n",
      "       \n",
      "        # propose phase\n",
      "        for woman in women:\n",
      "            ideal_man = woman.get_highest_ranked_person()\n",
      "            proposals[ideal_man].append(woman)\n",
      "       \n",
      "        # decision phase\n",
      "        for man in men:\n",
      "            if len(proposals[man]) == 0:\n",
      "                continue\n",
      "            best = proposals[man][0]\n",
      "            for proposer in proposals[man]:\n",
      "                if man.get_rank(proposer) > man.get_rank(best):\n",
      "                    proposer.remove_from_list(man) # reject\n",
      "                    has_rejection = True\n",
      "                if man.get_rank(proposer) < man.get_rank(best):\n",
      "                    best.remove_from_list(man) # reject\n",
      "                    has_rejection = True\n",
      "                    best = proposer\n",
      "            man.maybe(best)\n",
      "        num_days += 1\n",
      "    return num_days"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 63
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def print_women_pairing():\n",
      "    \"\"\"\n",
      "    Prints out the final pairing where women make the proposal\n",
      "    \"\"\"\n",
      "\n",
      "    print \"Final pairings:\"\n",
      "    print \"(Man, Woman)\"\n",
      "    for woman in women:\n",
      "        print \"(%s, %s)\" % (woman.get_partner().num, woman.num)"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 64
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def get_women_pairing():\n",
      "    \"\"\"\n",
      "    Retrieves the final pairing where women make the proposal\n",
      "    \"\"\"\n",
      "\n",
      "    pairings = []\n",
      "    for woman in women:\n",
      "        pairings.append((woman.get_partner().num, woman.num))\n",
      "    return sorted(pairings)  # sort by man's ID"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 65
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Test your implementation below."
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "women_propose()\n",
      "women_pairing = get_women_pairing()\n",
      "print_women_pairing()\n",
      "if women_pairing == [(1, 2), (2, 1), (3, 3)]:\n",
      "    print \"\\nCorrect implementation, nice job!\"\n",
      "else:\n",
      "    print \"\\nSomething is wrong with your implementation.\""
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "Final pairings:\n",
        "(Man, Woman)\n",
        "(2, 1)\n",
        "(1, 2)\n",
        "(3, 3)\n",
        "\n",
        "Correct implementation, nice job!\n"
       ]
      }
     ],
     "prompt_number": 66
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "<a id='q3'></a>\n",
      "## Question 3: Same final pairings"
     ]
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "This question will focus on exploring what happens when we randomize preferences. Below, you will find an implementation of the function `create_random_lists`, which allows you to create a random set of preferences. This function will be used to explore some questions about a typical case of the propose-and-reject algorithm."
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def create_random_lists(n=3):\n",
      "    \"\"\"\n",
      "    Create random lists of men and women with random preferences\n",
      "    \n",
      "    n -- the number of men/women, default to 3\n",
      "    \"\"\"\n",
      "\n",
      "    import random\n",
      "    people = list(range(1, n+1))\n",
      "    men = [Person('M', i, random.sample(people, n)) for i in range(1, n+1)]\n",
      "    women = [Person('W', i, random.sample(people, n)) for i in range(1, n+1)]\n",
      "    return men, women"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 67
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "If you run the cell below many times (repeated Ctrl+Enter), you will see that it prints different preference lists each time!"
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "men, women = create_random_lists(4)\n",
      "print ' | '.join(str(m) for m in men)\n",
      "print ' | '.join(str(w) for w in women)"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "output_type": "stream",
       "stream": "stdout",
       "text": [
        "M,1,[1, 4, 2, 3],None | M,2,[1, 2, 3, 4],None | M,3,[1, 3, 4, 2],None | M,4,[2, 3, 4, 1],None\n",
        "W,1,[2, 1, 3, 4],None | W,2,[3, 4, 2, 1],None | W,3,[4, 2, 3, 1],None | W,4,[4, 2, 1, 3],None\n"
       ]
      }
     ],
     "prompt_number": 68
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "We're interested in how often the men-propose and women-propose algorithms return the same pairing. Complete the following function that generates a random list of preferences for men and women, then runs each variant of the propose-and-reject algorithm on that set of preferences. Make sure to use the same set of preferences for each iteration of the algorithm.\n",
      "\n",
      "Follow the comments to carry out what you need to do for this question."
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def same_final_pairing(n=100, num_people=4):\n",
      "    \"\"\"\n",
      "    Runs the propose-and-reject on random preference lists n times,\n",
      "    and report the number of times men-propose and women-propose\n",
      "    give the same final pairing\n",
      "    \n",
      "    n -- number of iterations\n",
      "    num_people -- how large our instance is\n",
      "    \n",
      "    Please note that there are four (4) locations where you have to fill in\n",
      "    to make the function complete. At each location, you need at most two\n",
      "    lines of code.\n",
      "    \"\"\"\n",
      "\n",
      "    # Using global variables is highly discouraged in Python\n",
      "    # and programming in general, but for syntax simplicity,\n",
      "    # we'll be using it in this lab\n",
      "    global men, women\n",
      "    count = 0\n",
      "\n",
      "    # Iterate n times\n",
      "    for _ in range(n):\n",
      "        # Generates random list of preferences\n",
      "        # YOUR CODE HERE\n",
      "        men, women = create_random_lists(num_people)\n",
      "\n",
      "        # Runs the men-propose algorithm\n",
      "        # YOUR CODE HERE\n",
      "        men_propose()\n",
      "        men_pairing = get_men_pairing()\n",
      "    \n",
      "        # Resets the algorithm by removing the current pairings\n",
      "        # Do NOT edit the four lines below, or your implementation will be buggy\n",
      "        for man in men:\n",
      "            man.reset()\n",
      "        for woman in women:\n",
      "            woman.reset()\n",
      "\n",
      "        # Runs the women-propose algorithm\n",
      "        # YOUR CODE HERE\n",
      "        women_propose()\n",
      "        women_pairing = get_women_pairing()\n",
      "        \n",
      "        # Compares the final pairings, increments the counter\n",
      "        # if the pairings are the same\n",
      "        # YOUR CODE HERE\n",
      "        if men_pairing == women_pairing:\n",
      "            count += 1\n",
      "    return count"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 69
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Test your implementation below. How often is the male-optimal stable pairing the same as the female-optimal stable pairing? Does that frequency change as we change the value of `n`?\n",
      "\n",
      "Note that there is no one correct answer here - however, you should be sure to back up your statements with numbers (for instance, how many times you've seen each at some value of `n`)."
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "same_final_pairing()"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "metadata": {},
       "output_type": "pyout",
       "prompt_number": 70,
       "text": [
        "56"
       ]
      }
     ],
     "prompt_number": 70
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      " "
     ]
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "<a id='q4'></a>\n",
      "## Question 4: How many days does the algorithm take?"
     ]
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Finally, we're going to explore how long the men-propose algorithm takes as a function of `n`. Note that the function returns how many days it takes, so we can get the number of days taken with `num_days = men_propose()`."
     ]
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "Write a function that returns how many days it takes for the Propose and Reject algorithm, where women make the proposal, to arrive at its stable solution with a random preference list."
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "def how_long(num_people=10):\n",
      "    \"\"\"\n",
      "    Returns how many days the female-optimal algorithm takes\n",
      "    to produce the final pairings for randomly-generated preference lists\n",
      "    \"\"\"\n",
      "\n",
      "    global men, women\n",
      "    \n",
      "    # YOUR CODE HERE\n",
      "    men, women = create_random_lists(num_people)\n",
      "    return women_propose()"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [],
     "prompt_number": 72
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "how_long(200)"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "metadata": {},
       "output_type": "pyout",
       "prompt_number": 73,
       "text": [
        "246"
       ]
      }
     ],
     "prompt_number": 73
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "In lecture, you learned that the algorithm must terminate after at most $n^2$ days (you will soon prove a stricter bound in another problem in this homework).\n",
      "\n",
      "Try it on some inputs of different size. How quickly does the number of days grow with the input? Does it grow linearly (at the same rate as n - maybe twice as fast or half as fast), quadratically, logarithmically? If you can't tell, try graphing some outputs by hand. You do not have to submit any graph, but it certainly would help defend your claim."
     ]
    },
    {
     "cell_type": "code",
     "collapsed": false,
     "input": [
      "import matplotlib.pyplot as plt\n",
      "import numpy as np\n",
      "%matplotlib inline\n",
      "\n",
      "plt.plot(xrange(2, 200), map(how_long, xrange(2, 200)), '.')\n",
      "plt.xlabel('n -- number of people')\n",
      "plt.ylabel('how many days')\n",
      "plt.title(\"How many days the female-optimal algorithm takes as a function of n\")\n",
      "plt.show()"
     ],
     "language": "python",
     "metadata": {},
     "outputs": [
      {
       "metadata": {},
       "output_type": "display_data",
       "png": "iVBORw0KGgoAAAANSUhEUgAAAbQAAAEZCAYAAAD/ttB2AAAABHNCSVQICAgIfAhkiAAAAAlwSFlz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\n+Dia32l7AAAAAElFTkSuQmCC\n",
       "text": [
        "<matplotlib.figure.Figure at 0x9c7242c>"
       ]
      }
     ],
     "prompt_number": 27
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "**Congratulations**! You are done with Virtual Lab 3.\n",
      "\n",
      "Don't forget to convert this notebook to a pdf document, merge it with your written homework, and submit both the pdf and the code (as a zip file) on glookup.\n",
      "\n",
      "**Reminder**: late submissions are NOT accepted. If you have any technical difficulty,\n",
      "resolve it early on or use the provided VM."
     ]
    },
    {
     "cell_type": "markdown",
     "metadata": {},
     "source": [
      "**Acknowledgment**: The Person class was originally written by Mike Ambrose, one of our readers this semester."
     ]
    }
   ],
   "metadata": {}
  }
 ]
}